import os


def get_config(key: str, raise_exception: bool = True, default_value: str = None) -> str:
    """Retrieves the configuration value for a given key.

    Args:
        key (str): The configuration key to retrieve.
        raise_exception (bool, optional): Whether to raise an exception if the key is not found. Defaults to True.
        default_value (str, optional): The default value to return if the key is not found. Defaults to None.

    Raises:
        Exception: If the key is not found and raise_exception is True.

    Returns:
        str: The configuration value.
    """
    config = os.environ.get(key)
    if config:
        return config
    if raise_exception:
        raise Exception(f'Define the environment variable {key}')
    else:
        return default_value




from airless.core.hook import BaseHook


class LLMHook(BaseHook):

    def __init__(self):
        super().__init__()
        self.historic = ''

    def historic_append(self, text, actor):
        self.historic += f"---\n{actor}\n---\n{text}\n"

    def generate_completion(self, content, **kwargs):
        raise NotImplementedError()




from email import encoders
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication

from airless.core.hook import BaseHook


class EmailHook(BaseHook):
    """EmailHook class to build and send email messages.

    This class is responsible for constructing email messages that may 
    include attachments and other related information. However, the 
    sending functionality is not implemented.

    Inherits from:
        BaseHook: The base class for hooks in the airless framework.
    """

    def __init__(self):
        """Initializes the EmailHook class.

        This constructor calls the superclass constructor.
        """
        super().__init__()

    def build_message(self, subject: str, content: str, recipients: list, sender: str, attachments: list=[], mime_type: str='plain'):
        """Builds an email message with optional attachments.

        Args:
            subject (str): The subject of the email.
            content (str): The body content of the email.
            recipients (list): A list of recipient email addresses.
            sender (str): The email address of the sender.
            attachments (list, optional): A list of attachment dictionaries. 
                Each dictionary should contain 'name', 'content', and optionally 'type'. 
                Defaults to an empty list.
            mime_type (str, optional): The MIME type of the email body content. 
                Defaults to 'plain'.

        Returns:
            MIMEMultipart or MIMEText: The constructed email message object.
        """

        msg = MIMEText(content, mime_type)
        if attachments:
            msg = MIMEMultipart()
        msg['Subject'] = subject
        msg['To'] = ','.join(recipients)
        msg['From'] = sender

        for att in attachments:
            if att.get('type', 'text') == 'text':
                part = MIMEApplication(
                    att['content'],
                    Name=att['name']
                )
            else:
                part = MIMEBase('application', 'octet-stream')
                part.set_payload(att['content'])
                encoders.encode_base64(part)
            part['Content-Disposition'] = 'attachment; filename="%s"' % att['name']
            msg.attach(part)
        return msg

    def send(self, subject: str, content: str, recipients: list, sender: str, attachments: list, mime_type: str):
        """Sends the constructed email message.

        This method is not implemented and will raise a NotImplementedError.

        Args:
            subject (str): The subject of the email.
            content (str): The body content of the email.
            recipients (list): A list of recipient email addresses.
            sender (str): The email address of the sender.
            attachments (list): A list of attachment dictionaries.
            mime_type (str): The MIME type of the email body content.

        Raises:
            NotImplementedError: This method has not been implemented.
        """

        raise NotImplementedError()



from .base import (BaseHttpOperator, BaseFileOperator, BaseEventOperator)
from .delay import (DelayOperator)
from .error import (ErrorReprocessOperator)
from .redirect import (RedirectOperator)

__all__ = [
    'BaseHttpOperator',
    'BaseFileOperator',
    'BaseEventOperator',
    'DelayOperator',
    'ErrorReprocessOperator',
    'RedirectOperator'
]




from typing import Any

from airless.core.operator import BaseEventOperator


class RedirectOperator(BaseEventOperator):

    """
    Operator that receives one event from a queue topic and publishes multiple 
    messages to another topic.

    Parameters:
        event: The event received from the queue topic.
        project (str): The project where the destination queue is hosted.
        topic (str): The queue topic to which the newly generated messages will be published.
        messages (list): A list of messages to publish to the topic.
        params (list of dict): A list of dictionaries containing a key and a list of values.

    Returns:
        list: A list of output messages, which will be the product of the initial messages 
              and every list of values in params.
    """

    def __init__(self):
        """Initializes the RedirectOperator."""
        super().__init__()

    def execute(self, data: dict, topic: str):
        """
        Executes the operator, publishing messages to a specified topic.

        Args:
            data (dict): A dictionary containing event data with keys:
                - project (str): The project where the destination queue is hosted.
                - topic (str): The queue topic to which the newly generated messages will be published.
                - messages (list): A list of initial messages to publish.
                - params (list of dict): A list of parameters to modify the messages.
            topic (str): The topic to publish the messages to.

        Returns:
            None
        """

        to_project = data.get('project')
        to_topic = data['topic']
        messages = data.get('messages', [{}])
        params = data.get('params', [])

        messages = self.add_params_to_messages(messages, params)

        for msg in messages:
            self.queue_hook.publish(to_project, to_topic, msg)

    def add_params_to_messages(self, messages: list, params: list):
        """
        Adds parameters to each message in a list of messages.

        Args:
            messages (list): A list of messages to be modified.
            params (list of dict): A list of parameter dictionaries, each containing
                a key and a list of values.

        Returns:
            list: A list of messages with the parameters added.
        """

        for param in params:
            messages = self.add_param_to_messages(messages, param)
        return messages

    def add_param_to_messages(self, messages: list, param: dict):
        """
        Adds a single parameter to each message in a list of messages.

        Args:
            messages (list): A list of messages to modify.
            param (dict): A dictionary containing a 'key' and 'values'.

        Returns:
            list: A list of modified messages with the parameter added.
        """

        messages_with_param = []
        for message in messages:
            messages_with_param += self.add_param_to_message(message, param)
        return messages_with_param

    def add_param_to_message(self, message: dict, param: dict):
        """
        Adds a parameter's values to a single message.

        Args:
            message (dict): The original message to which parameters will be added.
            param (dict): A dictionary containing a 'key' and a list of 'values'.

        Returns:
            list: A list of new messages with the parameter added.
        """

        messages = []
        for value in param['values']:
            tmp_message = message.copy()
            keys = param['key'].split('.')
            tmp_message = self.add_key(tmp_message, keys, value)
            messages.append(tmp_message)
        return messages

    def add_key(self, obj: dict, keys: list, value: Any):
        """Adds a value to a nested dictionary at a specified key path.

        This method takes an object (dictionary), a list of keys representing 
        a path to a location in that dictionary, and a value to insert at 
        that location. If the specified nested keys do not exist in the 
        dictionary, they will be created.

        Args:
            obj (dict): The dictionary to which the keys and value will be added.
            keys (list): A list of keys representing the path in the dictionary.
            value: The value to be set at the specified key path.

        Returns:
            dict: A new dictionary with the value added at the specified key path.

        Example:
            obj = {'a': {'b': 1}}
            keys = ['a', 'c']
            value = 2
            result = add_key(self, obj, keys, value)
            # result will be {'a': {'b': 1, 'c': 2}}
        """

        tmp_obj = obj.copy()
        if len(keys) == 1:
            tmp_obj[keys[0]] = value
        else:
            nested_obj = tmp_obj.setdefault(keys[0], {})
            tmp_obj[keys[0]] = self.add_key(nested_obj, keys[1:], value)
        return tmp_obj




import json
import ndjson
import os
import requests
import re
import uuid

from datetime import datetime
from dateutil import parser
from ftplib import FTP
from typing import Any

from airless.core.hook import BaseHook


class FileHook(BaseHook):
    """FileHook class for handling file operations.

    This class provides methods to write data to local files in
    various formats (JSON and NDJSON), download files, rename files,
    and list files in a directory.

    Inherits from:
        BaseHook: The base class for hooks in the airless framework.
    """

    def __init__(self):
        """Initializes a new instance of the FileHook class."""
        super().__init__()

    def write(self, local_filepath: str, data: Any, **kwargs) -> None:
        """
        Writes data to a local file with support for JSON and NDJSON formats.

        Args:
            local_filepath (str):
                The path to the local file where the data will be written.
            data (Any):
                The data to write to the file. It can be a string, dictionary, list,
                or any other type that can be serialized to JSON or converted to a string.
        Kwargs:
            use_ndjson (bool):
                If `True` and the data is a dictionary or list, the data will be
                written in NDJSON format. Defaults to `False`.
            mode (str):
                The mode in which the file is opened. Common modes include:
                - `'w'`: Write mode, which overwrites the file if it exists.
                - `'wb'`: Write binary mode, which overwrites the file if it exists.
                Defaults to `'w'`.
        """

        use_ndjson = kwargs.get('use_ndjson', False)
        mode = kwargs.get('mode', 'w')

        with open(local_filepath, mode) as f:
            if mode == 'wb':
                f.write(data)
            elif isinstance(data, (dict, list)):
                dump = ndjson.dump if use_ndjson else json.dump
                dump(data, f)
            else:
                f.write(str(data))

    def extract_filename(self, filepath_or_url):
        """Extracts the filename from a filepath or URL.

        Args:
            filepath_or_url (str): The original file path or URL.

        Returns:
            str: The extracted filename.
        """
        return filepath_or_url.split('/')[-1].split('?')[0].split('#')[0]

    def get_tmp_filepath(self, filepath_or_url: str, **kwargs) -> str:
        """
        Generates a temporary file path based on the provided filepath or URL.

        Args:
            filepath_or_url (str):
                The original file path or URL from which the filename is extracted.

        Kwargs:
            add_timestamp (bool, optional):
                If `True`, a timestamp and a UUID will be prefixed to the filename to ensure uniqueness.
                Defaults to `True`.

        Returns:
            str: The temporary file path.
        """
        add_timestamp = kwargs.get('add_timestamp', True)

        filename = self.extract_filename(filepath_or_url)
        if add_timestamp:
            timestamp = datetime.now().strftime('%Y%m%d%H%M%S')
            filename = f"{timestamp}_{uuid.uuid4().hex}_{filename}"
        return f'/tmp/{filename}'

    def download(self, url, headers, timeout=500, proxies=None):
        """Downloads a file from a given URL and saves it to a temporary path.

        Args:
            url (str): The URL of the file to download.
            headers (dict): The headers to include in the request.
            timeout (int, optional): The request timeout in seconds. Defaults to 500.
            proxies (dict, optional): Proxy settings for the request. Defaults to None.

        Returns:
            str: The local filename where the downloaded file is saved.
        """

        local_filename = self.get_tmp_filepath(url)
        with requests.get(url, stream=True, verify=False, headers=headers, timeout=timeout, proxies=proxies) as r:
            r.raise_for_status()
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return local_filename

    def rename(self, from_filename, to_filename):
        """Renames a file from the original filename to the new filename.

        Args:
            from_filename (str): The original filename to rename.
            to_filename (str): The new filename.

        Returns:
            str: The new filename after renaming.
        """

        to_filename_formatted = ('' if to_filename.startswith('/tmp/') else '/tmp/') + to_filename
        os.rename(from_filename, to_filename_formatted)
        return to_filename_formatted

    def rename_files(self, dir, prefix):
        """Renames all files in a directory by prepending a prefix.

        Args:
            dir (str): The directory containing files to rename.
            prefix (str): The prefix to prepend to each file name.
        """

        for root, subdirs, files in os.walk(dir):
            for filename in files:
                os.rename(os.path.join(root, filename), os.path.join(root, f'{prefix}_{filename}'))

    def list_files(self, folder):
        """Lists all files in a specified directory.

        Args:
            folder (str): The folder path to search for files.

        Returns:
            list: A list of file paths found in the directory.
        """

        file_list = []
        for root, subdirs, files in os.walk(folder):
            for filename in files:
                filepath = os.path.join(root, filename)
                file_list.append(filepath)

        return file_list


class FtpHook(FileHook):
    """FtpHook class for handling FTP file operations.

    This class extends FileHook with methods specific to FTP file operations including
    connecting to an FTP server, navigating directories, and downloading files.
    """

    def __init__(self):
        """Initializes a new instance of the FtpHook class."""
        super().__init__()
        self.ftp = None

    def login(self, host, user, password):
        """Logs into the FTP server using the provided credentials.

        Args:
            host (str): The FTP server hostname or IP address.
            user (str): The username for the FTP server.
            password (str): The password for the FTP server.
        """

        self.ftp = FTP(host, user, password)
        self.ftp.login()

    def cwd(self, dir):
        """Changes the current working directory on the FTP server.

        Args:
            dir (str): The directory to change to.
        """

        if dir:
            self.ftp.cwd(dir)

    def dir(self):
        """Lists the files and directories in the current directory of the FTP server.

        This method retrieves a list of files and directories from the FTP server's 
        current working directory. It populates a list with the directory entries 
        and returns it.

        Returns:
            list: A list of directory entries as strings, each representing a file 
            or directory in the FTP server's current working directory.
        """
        lines = []
        self.ftp.dir("", lines.append)
        return lines


    def list(self, regex=None, updated_after=None, updated_before=None):
        """Lists files in the current directory of the FTP server with optional filters.

        Args:
            regex (str, optional): A regular expression to filter file names. Defaults to None.
            updated_after (datetime, optional): Filter files updated after this date. Defaults to None.
            updated_before (datetime, optional): Filter files updated before this date. Defaults to None.

        Returns:
            tuple: A tuple containing two lists:
                - A list of files (dictionaries with 'name' and 'updated_at').
                - A list of directories (dictionaries with 'name' and 'updated_at').
        """

        lines = self.dir()

        files = []
        directories = []

        for line in lines:
            tokens = line.split()
            obj = {
                'name': tokens[3],
                'updated_at': parser.parse(' '.join(tokens[:1]))
            }

            if regex and not re.search(regex, obj['name'], re.IGNORECASE):
                continue

            if updated_after and not (obj['updated_at'] >= updated_after):
                continue

            if updated_before and not (obj['updated_at'] <= updated_before):
                continue

            obj = {
                'name': tokens[3],
                'updated_at': parser.parse(' '.join(tokens[:1]))
            }
            if tokens[2] == '<DIR>':
                directories.append(obj)
            else:
                files.append(obj)

        return files, directories

    def download(self, dir, filename):
        """Downloads a file from the FTP server to a temporary local file.

        Args:
            dir (str): The directory on the FTP server where the file is located.
            filename (str): The name of the file to download.

        Returns:
            str: The local filepath where the downloaded file is saved.
        """

        self.cwd(dir)
        local_filepath = self.get_tmp_filepath(filename)
        with open(local_filepath, 'wb') as file:
            self.ftp.retrbinary(f'RETR {filename}', file.write)
        return local_filepath



import logging

from airless.core.utils import get_config


class BaseClass:
    """Base class for all components with logging capabilities."""

    def __init__(self) -> None:
        """Initializes the BaseClass and sets up logging."""
        self.logger = logging.getLogger(f'{self.__class__.__module__}.{self.__class__.__name__}')
        logging.basicConfig(level=logging.getLevelName(get_config('LOG_LEVEL')))
        self.logger.debug(f'Created class instance {self.__class__.__name__}')



from enum import Enum


class BaseEnum(Enum):
    """Base class for enumerations."""

    @classmethod
    def list(cls) -> list:
        """Lists all enumeration values.

        Returns:
            list: A list of enumeration values.
        """
        return list(map(lambda c: c, cls))

    @classmethod
    def find_by_id(cls, id: str) -> 'BaseEnum':
        """Finds an enumeration by its ID.

        Args:
            id (str): The ID to search for.

        Returns:
            BaseEnum: The found enumeration or None if not found.
        """
        return next(filter(lambda x: x == id, cls.list()), None)

    def __eq__(self, other) -> bool:
        """Checks equality with another object.

        Args:
            other (BaseEnum | dict | str): The object to compare with.

        Returns:
            bool: True if equal, False otherwise.
        """
        if isinstance(other, BaseEnum):
            return self.value['id'] == other.value['id']

        elif isinstance(other, dict):
            return self.value['id'] == other['id']

        else:
            return self.value['id'] == other



from .base import (BaseHook)
from .datalake import (DatalakeHook)
from .email import (EmailHook)
from .file import (FileHook, FtpHook)
from .queue import (QueueHook)
from .secret import (SecretManagerHook)
from .llm import (LLMHook)

__all__ = [
    'BaseHook',
    'DatalakeHook',
    'EmailHook',
    'FileHook',
    'FtpHook',
    'QueueHook',
    'SecretManagerHook',
    'LLMHook'
]



from airless.core import BaseClass


class BaseHook(BaseClass):
    """Base class for hooks in the system."""

    def __init__(self) -> None:
        """Initializes the BaseHook."""
        super().__init__()




import json
import time

from typing import Dict, Any

from airless.core.hook import DatalakeHook
from airless.core.operator import BaseEventOperator
from airless.core.utils import get_config


class ErrorReprocessOperator(BaseEventOperator):
    """Operator to handle processing of erroneous events.

    This operator manages the retry logic for events that fail.
    It can reprocess events based on configured retries and intervals,
    and if the maximum retries are exceeded, it saves the error
    details to the datalake
    """
    
    def __init__(self):
        """Initializes the ErrorReprocessOperator.
        
        Inherits from the BaseEventOperator and performs any necessary 
        initialization.  
        """
        super().__init__()
        self.datalake_hook = DatalakeHook()

    def execute(self, data, topic):
        """Executes the error processing logic for the given data.

        Args:
            data (dict): The event data that needs to be processed.
            topic (str): The topic from which the event is received.

        Raises:
            KeyError: If required keys are missing in the data dictionary.

        This method retrieves necessary metadata from the input data, 
        handles retries based on the specified parameters, and publishes 
        either the retried event back to the original topic or saves the
        error details to the datalake if maximum retries have
        been exceeded.
        """
        
        project = data.get('project', 'undefined')

        input_type = data['input_type']
        origin = data.get('origin', 'undefined')
        message_id = data.get('event_id')
        original_data = data['data']
        metadata = original_data.get('metadata', {})

        retry_interval = metadata.get('retry_interval', 5)
        retries = metadata.get('retries', 0)
        max_retries = metadata.get('max_retries', 2)
        max_interval = metadata.get('max_interval', 480)
        error_dataset = metadata.get('dataset')
        error_table = metadata.get('table')

        if (input_type == 'event') and (retries < max_retries):
            time.sleep(min(retry_interval ** retries, max_interval))
            original_data.setdefault('metadata', {})['retries'] = retries + 1
            self.queue_hook.publish(
                project=project,
                topic=origin,
                data=original_data)

        else:
            self.datalake_hook.send_to_landing_zone(
                data=data,
                dataset=error_dataset or get_config('ERROR_DATASET'),
                table=error_table or get_config('ERROR_TABLE'),
                message_id=message_id,
                origin=origin,
                time_partition=True)

            self._notify_email(origin=origin, message_id=message_id, data=data)
            self._notify_slack(origin=origin, message_id=message_id, data=data)

    def _notify_email(self, origin: str, message_id: str, data: Dict[str, Any]) -> None:
        """Sends an error notification to email if the env var `QUEUE_TOPIC_EMAIL_SEND` is defined

        Args:
            origin (str): The origin of the error.
            message_id (str): The ID of the message.
            data (Dict[str, Any]): The data related to the error.
        """
        email_send_topic = get_config('QUEUE_TOPIC_EMAIL_SEND', False)
        if email_send_topic and (origin != email_send_topic):
            email_message = {
                'sender': get_config('EMAIL_SENDER_ERROR'),
                'recipients': eval(get_config('EMAIL_RECIPIENTS_ERROR')),
                'subject': f'{origin} | {message_id}',
                'content': f'Input Type: {data["input_type"]} Origin: {origin}\nMessage ID: {message_id}\n\n {json.dumps(data["data"])}\n\n{data["error"]}'
            }
            self.queue_hook.publish(
                project=get_config('EMAIL_OPERATOR_PROJECT'),
                topic=email_send_topic,
                data=email_message)

    def _notify_slack(self, origin: str, message_id: str, data: Dict[str, Any]) -> None:
        """Sends an error notification to Slack if the env var `QUEUE_TOPIC_SLACK_SEND` is defined

        Args:
            origin (str): The origin of the error.
            message_id (str): The ID of the message.
            data (Dict[str, Any]): The data related to the error.
        """
        slack_send_topic = get_config('QUEUE_TOPIC_SLACK_SEND', False)
        if slack_send_topic and (origin != slack_send_topic):
            slack_message = {
                'channels': eval(get_config('SLACK_CHANNELS_ERROR')),
                'message': f'{origin} | {message_id}\n\n{json.dumps(data["data"])}\n\n{data["error"]}'
            }
            self.queue_hook.publish(
                project=get_config('SLACK_OPERATOR_PROJECT'),
                topic=slack_send_topic,
                data=slack_message)




import time

from airless.core.operator import BaseEventOperator


class DelayOperator(BaseEventOperator):

    """Operator that adds a delay to the pipeline.

    This operator introduces a delay in the processing pipeline by sleeping
    for a specified amount of time in seconds. The maximum delay that can
    be set is capped at 500 seconds.

    Attributes:
        None
    """

    def __init__(self):
        """Initializes a DelayOperator instance.

        This method calls the constructor of the parent class 
        `BaseEventOperator`.
        """

        super().__init__()

    def execute(self, data: dict, topic: str) -> None:

        """Executes the delay operation.

        Args:
            data (dict): A dictionary containing a key 'seconds' which
                determines how many seconds the operator should wait.
            topic: The topic to which the event is associated. This parameter
                is not utilized in this operator.

        Raises:
            KeyError: If the 'seconds' key is not present in the data dictionary.

        The function sleeps for the number of seconds specified, capping
        the maximum wait time at 500 seconds.
        """

        seconds = data['seconds']
        seconds = max(min(seconds, 500), 0)
        time.sleep(seconds)



from airless.core.hook import BaseHook


class QueueHook(BaseHook):
    """Hook for interacting with a queue system."""

    def __init__(self) -> None:
        """Initializes the QueueHook."""
        super().__init__()

    def publish(self, project: str, topic: str, data: dict) -> None:
        """Publishes data to a specified topic.

        Args:
            project (str): The project name.
            topic (str): The topic to publish to.
            data (dict): The data to publish.

        Raises:
            NotImplementedError: This method needs to be implemented in a subclass.
        """
        raise NotImplementedError()




import json

from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

from airless.core.hook import BaseHook
from airless.core.utils import get_config


class DatalakeHook(BaseHook):

    """DatalakeHook class is designer to write data to the datalake and must be
    implemented by a specific datalake vendor class

    Inherits from:
        BaseHook: The base class for hooks in the airless framework.
    """

    def __init__(self):
        """Initializes the DatalakeHook."""
        super().__init__()

    def build_metadata(self, message_id: Optional[int], origin: Optional[str]) -> Dict[str, Any]:
        """Builds metadata for the data being sent.

        Args:
            message_id (Optional[int]): The message ID.
            origin (Optional[str]): The origin of the data.

        Returns:
            Dict[str, Any]: The metadata dictionary.
        """
        return {
            'event_id': message_id or 1234,
            'resource': origin or 'local'
        }

    def prepare_row(self, row: Any, metadata: Dict[str, Any], now: datetime) -> Dict[str, Any]:
        """Prepares a row for insertion into the datalake.

        Args:
            row (Any): The row data.
            metadata (Dict[str, Any]): The metadata for the row.
            now (datetime): The current timestamp.

        Returns:
            Dict[str, Any]: The prepared row.
        """
        return {
            '_event_id': metadata['event_id'],
            '_resource': metadata['resource'],
            '_json': json.dumps({'data': row, 'metadata': metadata}),
            '_created_at': str(now)
        }

    def prepare_rows(self, data: Any, metadata: Dict[str, Any]) -> Tuple[List[Dict[str, Any]], datetime]:
        """Prepares multiple rows for insertion into the datalake.

        Args:
            data (Any): The data to prepare.
            metadata (Dict[str, Any]): The metadata for the rows.

        Returns:
            Tuple[List[Dict[str, Any]], datetime]: The prepared rows and the current timestamp.
        """
        now = datetime.now()
        prepared_rows = data if isinstance(data, list) else [data]
        return [self.prepare_row(row, metadata, now) for row in prepared_rows], now

    def send_to_landing_zone(self, data: Any, dataset: str, table: str, message_id: Optional[int], origin: Optional[str], time_partition: bool = False) -> Union[str, None]:
        """Sends data to the landing zone. This method must be implemented by the vendor specific class

        Args:
            data (Any): The data to send.
            dataset (str): The dataset name.
            table (str): The table name.
            message_id (Optional[int]): The message ID.
            origin (Optional[str]): The origin of the data.
            time_partition (bool, optional): Whether to use time partitioning. Defaults to False.

        Returns:
            Union[str, None]: The path to the uploaded file or None.
        """

        raise NotImplementedError('The vendor specific datalake class must implement this method')

    def _validate_non_empty_data(self, data, dataset, table):
        if isinstance(data, list) and (len(data) == 0):
            raise Exception(f'Trying to send empty list to landing zone: {dataset}.{table}')

        if isinstance(data, dict) and (data == {}):
            raise Exception(f'Trying to send empty dict to landing zone: {dataset}.{table}')

    def _dev_send_to_landing_zone(self, data, dataset, table):
        if get_config('ENV') == 'dev':
            self.logger.debug(f'[DEV] Uploading to {dataset}.{table}, Data: {data}')



from airless.core.service.base import BaseService


class CaptchaService(BaseService):
    """Service for solving captchas."""

    def __init__(self) -> None:
        """Initializes the CaptchaService."""
        super().__init__()

    def solve(self, version: str, key: str, url: str, action: str = 'verify') -> str:
        """Solves a captcha.

        Args:
            version (str): The version of the captcha (e.g., 'v2' or 'v3').
            key (str): The captcha key.
            url (str): The URL where the captcha is located.
            action (str, optional): The action to perform. Defaults to 'verify'.

        Raises:
            Exception: If the captcha version is not implemented.

        Returns:
            str: The solution to the captcha.
        """
        raise NotImplementedError()




import json
import time
import traceback

from base64 import b64decode

from airless.core import BaseClass
from airless.core.utils import get_config
from airless.core.hook import QueueHook


class BaseOperator(BaseClass):
    """BaseOperator class to handle message operations.

    This class provides the foundational functionality for various 
    operators that can handle different triggers, such as events, 
    files, and HTTP requests. It includes basic error handling 
    and message chaining functionalities.

    Inherits from:
        BaseClass: The base class for the operator implementations.
    """

    def __init__(self):
        """Initializes the BaseOperator class.

        Sets up necessary attributes for the operator including 
        a QueueHook instance for message publishing.
        """
        super().__init__()
        self.queue_hook = QueueHook()  # Have to redefine this attribute for each vendor
        self.trigger_type = None
        self.message_id = None
        self.has_error = False

    def extract_message_id(self, cloud_event):
        """Extracts the message ID from the cloud event.

        Args:
            cloud_event (CloudEvent): The cloud event from which to extract the message ID.

        Returns:
            str: The extracted message ID.
        """
        return cloud_event['id']

    def report_error(self, message: str, data: dict=None):
        """Reports an error by logging it and publishing to a queue.

        Args:
            message (str): The error message to report.
            data (dict, optional): Additional data associated with the error. Defaults to None.
        """
        if get_config('ENV') == 'prod':
            self.logger.error(f'Error {message}')
        else:
            self.logger.error(f'[DEV] Error {message}')

        error_obj = self.build_error_message(message, data)
        self.queue_hook.publish(
            project=None,
            topic=get_config('QUEUE_TOPIC_ERROR'),
            data=error_obj)

        self.has_error = True

    def build_error_message(self, message: str, data: dict):
        """Builds an error message.

        This method needs to be implemented in subclasses.

        Args:
            message (str): The error message.
            data (dict): The associated data.

        Raises:
            NotImplementedError: This method should be implemented by subclasses.
        """

        raise NotImplementedError()

    def chain_messages(self, messages: list) -> tuple:
        """Chains messages together for processing.

        Args:
            messages (list): A list of messages to chain.

        Returns:
            tuple: A tuple containing chained message data and the first topic.
        """

        msg_chain = None
        messages.reverse()

        for m in messages:
            new_msg = m['data'].copy()
            if msg_chain:
                new_msg['metadata'] = {**new_msg.get('metadata', {}), **msg_chain}

            msg_chain = {
                'run_next': [{
                    'topic': m['topic'],
                    'data': new_msg
                }]
            }

        chained_messages = msg_chain['run_next'][0]['data']
        first_topic = msg_chain['run_next'][0]['topic']

        return chained_messages, first_topic


class BaseFileOperator(BaseOperator):
    """BaseFileOperator class to handle file trigger operations.

    This class extends BaseOperator for operations triggered by files.
    """

    def __init__(self):
        """Initializes the BaseFileOperator class.

        Sets the trigger type to 'file' and initializes necessary attributes.
        """
        super().__init__()

        self.trigger_type = 'file'
        self.trigger_origin = None
        self.cloud_event = None

    def execute(self, bucket: str, filepath: str):
        """Executes file processing logic.

        This method needs to be implemented in subclasses.

        Args:
            bucket (str): The name of the bucket where the file is located.
            filepath (str): The path to the file within the bucket.

        Raises:
            NotImplementedError: This method should be implemented by subclasses.
        """
        raise NotImplementedError()

    def run(self, cloud_event):
        """Processes the incoming cloud event and executes file logic.

        Args:
            cloud_event (CloudEvent): The cloud event containing metadata about the file.
        """
        self.logger.debug(cloud_event)
        try:
            self.message_id = self.extract_message_id(cloud_event)
            self.cloud_event = cloud_event
            trigger_file_bucket = cloud_event['bucket']
            trigger_file_path = cloud_event.data['name']
            self.trigger_origin = f'{trigger_file_bucket}/{trigger_file_path}'
            self.execute(trigger_file_bucket, trigger_file_path)

        except Exception as e:
            self.report_error(f'{str(e)}\n{traceback.format_exc()}')

    def build_error_message(self, message: str, data: dict):
        """Builds an error message specific to file operations.

        Args:
            message (str): The error message.
            data (dict): The associated data.

        Returns:
            dict: A constructed error message.
        """
        return {
            'input_type': self.trigger_type,
            'origin': self.trigger_origin,
            'error': message,
            'event_id': self.message_id,
            'data': {
                'attributes': self.cloud_event._attributes,
                'data': data or self.cloud_event.data
            }
        }


class BaseEventOperator(BaseOperator):
    """BaseEventOperator class to handle event trigger operations.

    This class extends BaseOperator for operations triggered by events.
    """

    def __init__(self):
        """Initializes the BaseEventOperator class.

        Sets the trigger type to 'event' and initializes necessary attributes.
        """
        super().__init__()

        self.trigger_type = 'event'
        self.trigger_event_topic = None
        self.trigger_event_data = None

    def execute(self, data: dict, topic: str):
        """Executes event processing logic.

        This method needs to be implemented in subclasses.

        Args:
            data (dict): The data associated with the event.
            topic (str): The event topic.

        Raises:
            NotImplementedError: This method should be implemented by subclasses.
        """
        raise NotImplementedError()

    def run(self, cloud_event):
        """Processes the incoming cloud event and executes event logic.

        Args:
            cloud_event (CloudEvent): The cloud event containing metadata about the event.
        """

        self.logger.debug(cloud_event)
        try:
            self.message_id = self.extract_message_id(cloud_event)
            decoded_data = b64decode(cloud_event.data['message']['data']).decode('utf-8')
            self.trigger_event_data = json.loads(decoded_data)
            self.trigger_event_topic = cloud_event['source'].split('/')[-1]

            self.execute(self.trigger_event_data, self.trigger_event_topic)

            if not self.has_error:
                tasks = self.trigger_event_data.get('metadata', {}).get('run_next', [])
                self.run_next(tasks)

        except Exception as e:
            self.report_error(f'{str(e)}\n{traceback.format_exc()}')

    def run_next(self, tasks: list):
        """Executes the next tasks in the pipeline.

        Args:
            tasks (list): A list of tasks to execute next.
        """

        if tasks:
            time.sleep(10)
        for t in tasks:
            self.queue_hook.publish(
                project=t.get('project'),
                topic=t['topic'],
                data=t['data'])

    def build_error_message(self, message: str, data: dict):
        """Builds an error message specific to event operations.

        Args:
            message (str): The error message.
            data (dict): The associated data.

        Returns:
            dict: A constructed error message.
        """

        return {
            'input_type': self.trigger_type,
            'origin': self.trigger_event_topic,
            'error': message,
            'event_id': self.message_id,
            'data': data or self.trigger_event_data
        }


class BaseHttpOperator(BaseOperator):
    """BaseHttpOperator class to handle HTTP trigger operations.

    This class extends BaseOperator for operations triggered by HTTP requests.
    """

    def __init__(self):
        """Initializes the BaseHttpOperator class.

        Sets the trigger type to 'http' and initializes necessary attributes.
        """
        super().__init__()

        self.trigger_type = 'http'
        self.trigger_base_url = None
        self.trigger_request = None

    def execute(self, request):
        """Executes HTTP request processing logic.

        This method needs to be implemented in subclasses.

        Args:
            request (Request): The HTTP request object.

        Raises:
            NotImplementedError: This method should be implemented by subclasses.
        """
        raise NotImplementedError()

    def run(self, request):
        """Processes the incoming HTTP request and executes processing logic.

        Args:
            request (Request): The HTTP request object.
        """
        self.logger.debug(request)
        try:
            self.trigger_request = {
                'url': request.base_url,
                'method': request.method,
                'form': request.form.to_dict(),
                'args': request.args.to_dict(),
                'data': request.data.decode('utf-8')
            }
            self.trigger_base_url = request.base_url

            return self.execute(request)

        except Exception as e:
            self.report_error(f'{str(e)}\n{traceback.format_exc()}')

    def build_error_message(self, message, request):
        """Builds an error message specific to HTTP operations.

        Args:
            message (str): The error message.
            request (Request): The HTTP request object.

        Returns:
            dict: A constructed error message.
        """
        return {
            'input_type': self.trigger_type,
            'origin': self.trigger_base_url,
            'error': message,
            'event_id': int(time.time() * 1000),
            'data': request or self.trigger_request
        }



from airless.core import BaseClass


class BaseService(BaseClass):
    """Base class for services in the system."""

    def __init__(self) -> None:
        """Initializes the BaseService."""
        super().__init__()
